@@ -4,6 +4,7 @@ import uuid
 import os
 import io
 
+from multiprocessing import Process
 from pyroute2.common import uuid32
 from pyroute2.common import debug
 from pyroute2.netlink import NLMSG_CONTROL
@@ -20,6 +21,7 @@ from pyroute2.netlink import IPRCMD_SUBSCRIBE
 from pyroute2.netlink import IPRCMD_PROVIDE
 from pyroute2.netlink import IPRCMD_REMOVE
 from pyroute2.netlink import IPRCMD_DISCOVER
+from pyroute2.netlink import IPRCMD_STOP
 from pyroute2.netlink import NLMSG_DONE
 from pyroute2.netlink import NLM_F_MULTI
 from pyroute2.netlink.generic import mgmtmsg
@@ -29,6 +31,7 @@ from pyroute2.iocore import NLT_RESPONSE
 from pyroute2.iocore import NLT_EXCEPTION
 from pyroute2.iocore.loop import IOLoop
 from pyroute2.iocore.broker import pairPipeSockets
+from pyroute2.iocore.broker import MarshalEnv
 from pyroute2.iocore.broker import IOBroker
 from pyroute2.iocore.addrpool import AddrPool
 
@@ -52,7 +55,7 @@ class IOCore(object):
 
     def __init__(self, debug=False, timeout=3, do_connect=False,
                  host=None, key=None, cert=None, ca=None,
-                 addr=None):
+                 addr=None, fork=False, secret=None):
         addr = addr or uuid32()
         self._timeout = timeout
         self.default_broker = addr
@@ -62,7 +65,9 @@ class IOCore(object):
         self.callbacks = []     # [(predicate, callback, args), ...]
         self.debug = debug
         self.cid = None
-        self.nonce = AddrPool()
+        self.cmd_nonce = AddrPool(minaddr=0xf, maxaddr=0xfffe)
+        self.nonce = AddrPool(minaddr=0xffff, maxaddr=0xffffffff)
+        self.emarshal = MarshalEnv()
         self.save = None
         if self.marshal is not None:
             self.marshal.debug = debug
@@ -73,14 +78,34 @@ class IOCore(object):
 
         self.ioloop = IOLoop()
 
-        self.iobroker = IOBroker(addr=addr, ioloop=self.ioloop)
-        self.iobroker.start()
         self._brs, self.bridge = pairPipeSockets()
-        self.iobroker.add_client(self._brs)
-        self.iobroker.controls.add(self._brs)
-        self.ioloop.register(self._brs,
-                             self.iobroker.route,
-                             defer=True)
+        # To fork or not to fork?
+        #
+        # It depends on how you gonna use RT netlink sockets
+        # in your application. Performance can also differ,
+        # and sometimes forked broker can even speedup your
+        # application -- keep in mind Python's GIL
+        if fork:
+            # Start the I/O broker in a separate process,
+            # so you can use multiple RT netlink sockets in
+            # one application -- it does matter, if your
+            # application already uses some RT netlink
+            # library and you want to smoothly try and
+            # migrate to the pyroute2
+            self.forked_broker = Process(target=self._start_broker,
+                                         args=(fork, secret))
+            self.forked_broker.start()
+        else:
+            # Start I/O broker as an embedded object, so
+            # the RT netlink socket will be opened in the
+            # same process. Technically, you can open
+            # multiple RT netlink sockets within one process,
+            # but only the first one will receive all the
+            # answers -- so effectively only one socket per
+            # process can be used.
+            self.forked_broker = None
+            self._start_broker(fork, secret)
+        self.ioloop.start()
         self.ioloop.register(self.bridge,
                              self._route,
                              defer=True)
@@ -92,47 +117,23 @@ class IOCore(object):
             self.default_dport = self.discover(self.default_target or path,
                                                self.default_peer)
 
+    def _start_broker(self, fork=False, secret=None):
+        iobroker = IOBroker(addr=self.default_broker,
+                            ioloop=None if fork else self.ioloop,
+                            control=self._brs, secret=secret)
+        iobroker.start()
+        if fork:
+            iobroker._stop_event.wait()
+
     @debug
     def _route(self, sock, raw):
-        buf = io.BytesIO()
-        buf.length = buf.write(raw)
-        buf.seek(0)
+        data = io.BytesIO()
+        data.length = data.write(raw)
 
-        if self.save is not None:
-            # concatenate buffers
-            buf.seek(0)
-            self.save.write(buf.read())
-            self.save.length += buf.length
-            # discard save
-            buf = self.save
-            self.save = None
+        for envelope in self.emarshal.parse(data, sock):
 
-        offset = 0
-        while offset < buf.length:
-            buf.seek(offset)
-            (length,
-             mtype,
-             flags) = struct.unpack('IHH', buf.read(8))
-
-            if offset + length > buf.length:
-                # create save buffer
-                buf.seek(offset)
-                self.save = io.BytesIO()
-                self.save.length = self.save.write(buf.read())
-                # truncate the buffer
-                buf.truncate(offset)
-                break
-
-            buf.seek(offset)
-            data = io.BytesIO()
-            data.write(buf.read(length))
-            data.length = length
-            data.seek(0)
-
-            # data traffic
-            envelope = envmsg(data)
-            envelope.decode()
             nonce = envelope['header']['sequence_number']
+            flags = envelope['header']['flags']
             try:
                 buf = io.BytesIO()
                 buf.length = buf.write(envelope.
@@ -149,8 +150,6 @@ class IOCore(object):
                 # now silently drop bad packet
                 pass
 
-            offset += length
-
     @debug
     def parse(self, envelope, data):
 
@@ -224,8 +223,10 @@ class IOCore(object):
         msg.encode()
         rsp = self.request(msg.buf.getvalue(),
                            env_flags=NLT_CONTROL,
+                           nonce_pool=self.cmd_nonce,
                            addr=addr)[0]
-        assert rsp['cmd'] == IPRCMD_ACK
+        if rsp['cmd'] != IPRCMD_ACK:
+            raise RuntimeError(rsp.get_attr('IPR_ATTR_ERROR'))
         if expect is not None:
             if type(expect) not in (list, tuple):
                 return rsp.get_attr(expect)
@@ -303,7 +304,9 @@ class IOCore(object):
             except Queue.Empty as e:
                 if addr == self.default_broker:
                     raise e
-        self.iobroker.shutdown()
+        self.command(IPRCMD_STOP)
+        if self.forked_broker:
+            self.forked_broker.join()
         self.ioloop.shutdown()
         self.ioloop.join()
 
@@ -394,12 +397,14 @@ class IOCore(object):
                 return
 
     @debug
-    def get(self, key=0, raw=False, timeout=None, terminate=None):
+    def get(self, key=0, raw=False, timeout=None, terminate=None,
+            nonce_pool=None):
         '''
         Get a message from a queue
 
         * key -- message queue number
         '''
+        nonce_pool = nonce_pool or self.nonce
         queue = self.listeners[key]
         result = []
         e = None
@@ -446,7 +451,7 @@ class IOCore(object):
         if key != 0:
             # delete the queue
             del self.listeners[key]
-            self.nonce.free(key)
+            nonce_pool.free(key)
             # get remaining messages from the queue and
             # re-route them to queue 0 or drop
             while not queue.empty():
@@ -487,14 +492,17 @@ class IOCore(object):
                 addr=None,
                 port=None,
                 nonce=None,
+                nonce_pool=None,
                 cname=None,
                 response_timeout=None,
                 terminate=None):
-        nonce = nonce or self.nonce.alloc()
+        nonce_pool = nonce_pool or self.nonce
+        nonce = nonce or nonce_pool.alloc()
         port = port or self.default_dport
         addr = addr or self.default_broker
         self.listeners[nonce] = Queue.Queue(maxsize=_QUEUE_MAXSIZE)
         self.push((addr, port), msg, env_flags, nonce, cname)
         return self.get(nonce,
+                        nonce_pool=nonce_pool,
                         timeout=response_timeout,
                         terminate=terminate)
