@@ -4,6 +4,7 @@ import socket
 import os
 import io
 
+from pyroute2.iocore.addrpool import AddrPool  # FIXME: move to common
 from pyroute2.netlink.generic import nlmsg
 from pyroute2.netlink.generic import NetlinkDecodeError
 from pyroute2.netlink.generic import NetlinkHeaderDecodeError
@@ -83,28 +84,51 @@ class Marshal(object):
         # one marshal instance can be used to parse one
         # message at once
         self.msg_map = self.msg_map or {}
+        self.defragmentation = {}
 
-    def parse(self, data):
+    def parse(self, data, sock=None):
         '''
         Parse the data in the buffer
+
+        If socket is provided, support defragmentation
         '''
         with self.lock:
-            total = data.length
             data.seek(0)
             offset = 0
             result = []
 
-            while offset < total:
+            if sock in self.defragmentation:
+                save = self.defragmentation[sock]
+                save.write(data.read())
+                save.length += data.length
+                # discard save
+                data = save
+                del self.defragmentation[sock]
+                data.seek(0)
+
+            while offset < data.length:
                 # pick type and length
                 (length, msg_type) = struct.unpack('IH', data.read(6))
+                data.seek(offset)
+                # if length + offset is greater than
+                # remaining size, save the buffer for
+                # defragmentation
+                if (sock is not None) and (length + offset > data.length):
+                    # create save buffer
+                    self.defragmentation[sock] = save = io.BytesIO()
+                    save.length = save.write(data.read())
+                    # truncate data
+                    data.truncate(offset)
+                    break
+
                 error = None
                 if msg_type == NLMSG_ERROR:
                     data.seek(offset + 16)
                     code = abs(struct.unpack('i', data.read(4))[0])
                     if code > 0:
                         error = NetlinkError(code)
+                    data.seek(offset)
 
-                data.seek(offset)
                 msg_class = self.msg_map.get(msg_type, nlmsg)
                 msg = msg_class(data, debug=self.debug)
                 try:
@@ -130,23 +154,70 @@ class Marshal(object):
         pass
 
 
+# 8<-----------------------------------------------------------
+# Singleton, containing possible modifiers to the NetlinkSocket
+# bind() call.
+#
+# Normally, you can open only one netlink connection for one
+# process, but there is a hack. Current PID_MAX_LIMIT is 2^22,
+# so we can use the rest to midify pid field.
+#
+# See also libnl library, lib/socket.c:generate_local_port()
+sockets = AddrPool(minaddr=0x0,
+                   maxaddr=0x3ff,
+                   reverse=True)
+# 8<-----------------------------------------------------------
+
+
 class NetlinkSocket(socket.socket):
     '''
     Generic netlink socket
     '''
 
-    def __init__(self, family=NETLINK_GENERIC):
+    def __init__(self, family=NETLINK_GENERIC, port=None):
         socket.socket.__init__(self, socket.AF_NETLINK,
                                socket.SOCK_DGRAM, family)
-        self.pid = os.getpid()
+        global sockets
+        self.pid = os.getpid() & 0x3fffff
+        self.port = port
+        self.fixed = self.port is not None
         self.groups = 0
         self.marshal = None
 
     def bind(self, groups=0):
         self.groups = groups
-        socket.socket.bind(self, (self.pid, self.groups))
+        # if we have pre-defined port, use it strictly
+        if self.fixed:
+            socket.socket.bind(self, (self.pid + (self.port << 22),
+                               self.groups))
+            return
+
+        # if we have no pre-defined port, scan all the
+        # range till the first available port
+        for i in range(1024):
+            try:
+                self.port = sockets.alloc()
+                socket.socket.bind(self,
+                                   (self.pid + (self.port << 22),
+                                    self.groups))
+                # if we're here, bind() done successfully, just exit
+                return
+            except socket.error as e:
+                # pass occupied sockets, raise other exceptions
+                if e.errno != 98:
+                    raise
+        else:
+            # raise "address in use" -- to be compatible
+            raise socket.error(98, 'Address already in use')
 
     def get(self):
         data = io.BytesIO()
         data.length = data.write(self.recv(16384))
         return self.marshal.parse(data)
+
+    def close(self):
+        global sockets
+        assert self.port is not None
+        if not self.fixed:
+            sockets.free(self.port)
+        socket.socket.close(self)
