@@ -679,7 +679,8 @@ class mgmtmsg(ctrlmsg):
                ('IPR_ATTR_ERROR', 'asciiz'),
                ('IPR_ATTR_CID', 'uint32'),
                ('IPR_ATTR_KEY', 'u32key'),
-               ('IPR_ATTR_UUID', 'asciiz'))
+               ('IPR_ATTR_UUID', 'asciiz'),
+               ('IPR_ATTR_SOURCE', 'uint32'))
 
     class u32key(nla):
         fields = (('offset', 'I'),
