@@ -5,10 +5,11 @@ unless you understand what are you doing.
 Quick start:
     from pyroute2 import IPDB
     ip = IPDB()
-    ip['eth0']['address'] = '00:11:22:33:44:55'
-    ip['eth0']['ifname'] = 'bala'
-    ip['eth0']['txqlen'] = 2000
-    ip['eth0'].commit()
+    ip.interfaces['eth0']['address'] = '00:11:22:33:44:55'
+    ip.interfaces['eth0']['ifname'] = 'bala'
+    ip.interfaces['eth0']['txqlen'] = 2000
+    ip.interfaces['eth0'].commit()
+    ip.routes.add({'dst': 'default', 'gateway': '10.0.0.1'}).commit()
 '''
 import uuid
 import platform
@@ -18,21 +19,17 @@ try:
 except ImportError:
     from queue import Empty
 
+from socket import AF_UNSPEC
 from socket import AF_INET
 from socket import AF_INET6
 from pyroute2.common import Dotkeys
 from pyroute2.netlink import NetlinkError
 from pyroute2.netlink.iproute import IPRoute
+from pyroute2.netlink.rtnl.rtmsg import rtmsg
 from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
 from pyroute2.netlink.rtnl.tcmsg import tcmsg
 
 tc_fields = [tcmsg.nla2name(i[0]) for i in tcmsg.nla_map]
-nla_fields = [ifinfmsg.nla2name(i[0]) for i in ifinfmsg.nla_map]
-nla_fields.append('flags')
-nla_fields.append('mask')
-nla_fields.append('change')
-nla_fields.append('state')
-nla_fields.append('removal')
 
 
 _ANCIENT_PLATFORM = platform.dist()[:2] == ('redhat', '6.4')
@@ -195,17 +192,39 @@ def update(f):
     return decorated
 
 
-class IPLinkRequest(dict):
+class IPRequest(dict):
 
-    def __init__(self, interface=None):
+    def __init__(self, obj=None):
         dict.__init__(self)
-        if interface:
-            self.update(interface)
+        if obj is not None:
+            self.update(obj)
 
-    def update(self, interface):
-        for key in interface:
-            if interface[key] is not None:
-                self[key] = interface[key]
+    def update(self, obj):
+        for key in obj:
+            if obj[key] is not None:
+                self[key] = obj[key]
+
+
+class IPRouteRequest(IPRequest):
+
+    def __setitem__(self, key, value):
+        if (key == 'dst') and (value != 'default'):
+            value = value.split('/')
+            if len(value) == 1:
+                dst = value[0]
+                mask = 0
+            elif len(value) == 2:
+                dst = value[0]
+                mask = int(value[1])
+            else:
+                raise ValueError('wrong destination')
+            dict.__setitem__(self, 'dst', dst)
+            dict.__setitem__(self, 'dst_len', mask)
+        elif key != 'dst':
+            dict.__setitem__(self, key, value)
+
+
+class IPLinkRequest(IPRequest):
 
     def __setitem__(self, key, value):
         if key == 'kind':
@@ -228,8 +247,8 @@ class Transactional(Dotkeys):
     '''
     An utility class that implements common transactional logic.
     '''
-    def __init__(self, nl=None, ipdb=None, mode='direct'):
-        self.nl = nl
+    def __init__(self, ipdb, mode=None):
+        self.nl = ipdb.nl
         self.uid = uuid.uuid4()
         self.ipdb = ipdb
         self.last_error = None
@@ -237,7 +256,7 @@ class Transactional(Dotkeys):
         self._fields = []
         self._tids = []
         self._transactions = {}
-        self._mode = mode
+        self._mode = mode or ipdb.mode
         self._write_lock = threading.RLock()
         self._direct_state = State(self._write_lock)
         self._linked_sets = set()
@@ -262,13 +281,13 @@ class Transactional(Dotkeys):
         used as transactions.
         '''
         with self._write_lock:
-            res = self.__class__(nl=self.nl, mode='snapshot')
+            res = self.__class__(ipdb=self.ipdb, mode='snapshot')
             for key in tuple(self.keys()):
-                if key in nla_fields:
+                if key in self._fields:
                     res[key] = self[key]
             for key in self._linked_sets:
                 res[key] = LinkedSet(self[key])
-                if self.ipdb is not None and not detached:
+                if not detached:
                     self[key].connect(res[key])
             return res
 
@@ -299,7 +318,7 @@ class Transactional(Dotkeys):
     def __sub__(self, vs):
         '''
         '''
-        res = self.__class__(nl=self.nl, mode='snapshot')
+        res = self.__class__(ipdb=self.ipdb, mode='snapshot')
         with self._direct_state:
             # simple keys
             for key in self:
@@ -321,10 +340,13 @@ class Transactional(Dotkeys):
         '''
         # keep snapshot's ip addr set updated from the OS
         # it is required by the commit logic
-        t = self.pick(detached=False)
-        self._transactions[t.uid] = t
-        self._tids.append(t.uid)
-        return t.uid
+        with self.ipdb.exclusive:
+            if self.ipdb._stop:
+                raise RuntimeError("Can't start transaction on released IPDB")
+            t = self.pick(detached=False)
+            self._transactions[t.uid] = t
+            self._tids.append(t.uid)
+            return t.uid
 
     def last(self):
         '''
@@ -385,8 +407,8 @@ class Transactional(Dotkeys):
 class TrafficControl(Transactional):
     '''
     '''
-    def __init__(self, nl=None, ipdb=None, mode='direct'):
-        Transactional.__init__(self, nl, ipdb, mode)
+    def __init__(self, ipdb, mode=None):
+        Transactional.__init__(self, ipdb, mode)
         self._fields = tc_fields
 
     def load(self, msg):
@@ -425,7 +447,7 @@ class Interface(Transactional):
     exception will be raised. Failed transaction review
     will be attached to the exception.
     '''
-    def __init__(self, nl=None, ipdb=None, mode='direct'):
+    def __init__(self, ipdb, mode=None):
         '''
         One can use interface objects standalone as
         well as in connection with ipdb object. Standalone
@@ -436,7 +458,7 @@ class Interface(Transactional):
             * ipdb -- ipdb() reference
             * mode -- transaction mode
         '''
-        Transactional.__init__(self, nl, ipdb, mode)
+        Transactional.__init__(self, ipdb, mode)
         self.cleanup = ('header',
                         'linkinfo',
                         'af_spec',
@@ -448,7 +470,12 @@ class Interface(Transactional):
         self.ingress = None
         self.egress = None
         self._exists = False
-        self._fields = nla_fields
+        self._fields = [ifinfmsg.nla2name(i[0]) for i in ifinfmsg.nla_map]
+        self._fields.append('flags')
+        self._fields.append('mask')
+        self._fields.append('change')
+        self._fields.append('state')
+        self._fields.append('removal')
         self._load_event = threading.Event()
         self._linked_sets.add('ipaddr')
         self._linked_sets.add('ports')
@@ -457,7 +484,7 @@ class Interface(Transactional):
         with self._direct_state:
             self['ipaddr'] = LinkedSet()
             self['ports'] = LinkedSet()
-            for i in nla_fields:
+            for i in self._fields:
                 self[i] = None
             for i in ('state', 'change', 'mask'):
                 del self[i]
@@ -485,10 +512,8 @@ class Interface(Transactional):
         with self._direct_state:
             self._exists = True
             self.update(dev)
-            self._attrs = set()
             for (name, value) in dev['attrs']:
                 norm = ifinfmsg.nla2name(name)
-                self._attrs.add(norm)
                 self[norm] = value
             # load interface kind
             linkinfo = dev.get_attr('IFLA_LINKINFO')
@@ -562,7 +587,7 @@ class Interface(Transactional):
                 transaction.del_port(port)
         else:
             # FIXME: do something with it, please
-            if self.ipdb and _ANCIENT_PLATFORM and \
+            if _ANCIENT_PLATFORM and \
                     'master' in self.ipdb.interfaces[port]:
                 self.ipdb.interfaces[port].del_item('master')
             self['ports'].remove(port)
@@ -669,7 +694,7 @@ class Interface(Transactional):
             # Interface changes
             request = IPLinkRequest()
             for key in added:
-                if key in nla_fields:
+                if key in self._fields:
                     request[key] = added[key]
 
             # apply changes only if there is something to apply
@@ -792,6 +817,181 @@ class Interface(Transactional):
         self['removal'] = True
 
 
+class Route(Transactional):
+
+    def __init__(self, ipdb, mode=None):
+        Transactional.__init__(self, ipdb, mode)
+        self._exists = False
+        self._load_event = threading.Event()
+        self._fields = [rtmsg.nla2name(i[0]) for i in rtmsg.nla_map]
+        self._fields.append('flags')
+        self._fields.append('src_len')
+        self._fields.append('dst_len')
+        self._fields.append('table')
+        self._fields.append('removal')
+        self.cleanup = ('attrs',
+                        'header',
+                        'event')
+
+    def load(self, msg):
+        with self._direct_state:
+            self._exists = True
+            self.update(msg)
+            # merge key
+            for (name, value) in msg['attrs']:
+                norm = rtmsg.nla2name(name)
+                self[norm] = value
+            if msg.get_attr('RTA_DST', None) is not None:
+                dst = '%s/%s' % (msg.get_attr('RTA_DST'),
+                                 msg['dst_len'])
+            else:
+                dst = 'default'
+            self['dst'] = dst
+            # finally, cleanup all not needed
+            for item in self.cleanup:
+                if item in self:
+                    del self[item]
+
+            self.sync()
+
+    def sync(self):
+        self._load_event.set()
+
+    def reload(self):
+        # do NOT call get_routes() here, it can cause race condition
+        self._load_event.wait()
+
+    def commit(self, tid=None, transaction=None, rollback=False):
+        self._load_event.clear()
+        error = None
+
+        if tid:
+            transaction = self._transactions[tid]
+        else:
+            transaction = transaction or self.last()
+
+        # create a new route
+        if not self._exists:
+            try:
+                self.nl.route('add', **IPRouteRequest(self))
+            except Exception:
+                self.nl = None
+                self.ipdb.routes.remove(self)
+                raise
+
+        # work on existing route
+        snapshot = self.pick()
+        try:
+            # route set
+            request = IPRouteRequest(transaction - snapshot)
+            if any([request[x] is not None for x in request]):
+                self.nl.route('set', **IPRouteRequest(transaction))
+
+            if transaction.get('removal'):
+                self.nl.route('delete', **IPRouteRequest(snapshot))
+
+        except Exception as e:
+            if not rollback:
+                ret = self.commit(transaction=snapshot, rollback=True)
+                if ret is not None:
+                    error = ret
+                else:
+                    error = e
+            else:
+                self.drop()
+                x = RuntimeError()
+                x.cause = e
+                raise x
+
+        if not rollback:
+            self.drop()
+            self.reload()
+
+        if error is not None:
+            error.transaction = transaction
+            raise error
+
+    def remove(self):
+        self['removal'] = True
+
+
+class RoutingTables(dict):
+
+    def __init__(self, ipdb):
+        dict.__init__(self)
+        self.ipdb = ipdb
+        self.tables = dict()
+
+    def add(self, spec):
+        '''
+        Create a route from a dictionary
+        '''
+        table = spec.get('table', 254)
+        assert 'dst' in spec
+        route = Route(self.ipdb)
+        route.update(spec)
+        if table not in self.tables:
+            self.tables[table] = dict()
+        self.tables[table][route['dst']] = route
+        route.begin()
+        return route
+
+    def load(self, msg):
+        '''
+        Loads an existing route from a rtmsg
+        '''
+        table = msg.get('table', 254)
+        if table not in self.tables:
+            self.tables[table] = dict()
+
+        dst = msg.get_attr('RTA_DST', None)
+        if dst is None:
+            key = 'default'
+        else:
+            key = '%s/%s' % (dst, msg.get('dst_len', 0))
+
+        if key in self.tables[table]:
+            ret = self.tables[table][key]
+            ret.load(msg)
+        else:
+            ret = Route(ipdb=self.ipdb)
+            ret.load(msg)
+            self.tables[table][key] = ret
+        return ret
+
+    def remove(self, route, table=None):
+        if isinstance(route, Route):
+            table = route.get('table', 254)
+            route = route.get('dst', 'default')
+        else:
+            table = table or 254
+        del self.tables[table][route]
+
+    def get(self, dst, table=None):
+        table = table or 254
+        return self.tables[table][dst]
+
+    def keys(self, table=254, family=AF_UNSPEC):
+        return [x['dst'] for x in self.tables[table].values()
+                if (x['family'] == family) or (family == AF_UNSPEC)]
+
+    def has_key(self, key, table=254):
+        return key in self.tables[table]
+
+    def __contains__(self, key):
+        return key in self.tables[254]
+
+    def __getitem__(self, key):
+        return self.get(key)
+
+    def __setitem__(self, key, value):
+        assert key == value['dst']
+        return self.add(value)
+
+    def __delitem__(self, key):
+        return self.remove(key)
+
+
 class IPDB(object):
     '''
     The class that maintains information about network setup
@@ -800,7 +1000,8 @@ class IPDB(object):
     '''
 
     def __init__(self, nl=None, host=None, mode='implicit',
-                 key=None, cert=None, ca=None, iclass=Interface):
+                 key=None, cert=None, ca=None, iclass=Interface,
+                 fork=False):
         '''
         Parameters:
             * nl -- IPRoute() reference
@@ -811,21 +1012,27 @@ class IPDB(object):
         you can start two and more iproute instances, but
         only the first one will receive anything.
         '''
-        self.nl = nl or IPRoute(host=host, key=key, cert=cert, ca=ca)
+        self.nl = nl or IPRoute(host=host,
+                                key=key,
+                                cert=cert,
+                                ca=ca,
+                                fork=fork)
         self.mode = mode
         self.iclass = iclass
         self._stop = False
-        self._callbacks = []
+        # see also 'register_callback'
+        self._post_callbacks = []
+        self._pre_callbacks = []
         self._cb_threads = set()
 
         # resolvers
         self.interfaces = Dotkeys()
+        self.routes = RoutingTables(ipdb=self)
         self.by_name = Dotkeys()
         self.by_index = Dotkeys()
 
         # caches
         self.ipaddr = {}
-        self.routes = {}
         self.neighbors = {}
         self.old_names = {}
 
@@ -838,6 +1045,8 @@ class IPDB(object):
         self.update_links(links)
         self.update_slaves(links)
         self.update_addr(self.nl.get_addr())
+        routes = self.nl.get_routes()
+        self.update_routes(routes)
 
         # start monitoring thread
         self.nl.mirror()
@@ -851,28 +1060,85 @@ class IPDB(object):
     def __exit__(self, exc_type, exc_value, traceback):
         self.release()
 
-    #def __dir__(self):
-    #    ret = Dotkeys.__dir__(self)
-    #    ret.append('by_name')
-    #    ret.append('by_index')
-    #    return ret
+    def register_callback(self, callback, mode='post'):
+        '''
+        IPDB callbacks are routines executed on a RT netlink
+        message arrival. There are two types of callbacks:
+        "post" and "pre" callbacks.
 
-    def register_callback(self, callback):
-        self._callbacks.append(callback)
+        ...
 
-    def unregister_callback(self, callback):
-        for cb in tuple(self._callbacks):
+        "Post" callbacks are executed after the message is
+        processed by IPDB and all corresponding objects are
+        created or deleted. Using ipdb reference in "post"
+        callbacks you will access the most up-to-date state
+        of the IP database.
+
+        "Post" callbacks are executed asynchronously in
+        separate threads. These threads can work as long
+        as you want them to. Callback threads are joined
+        occasionally, so for a short time there can exist
+        stopped threads.
+
+        ...
+
+        "Pre" callbacks are synchronous routines, executed
+        before the message gets processed by IPDB. It gives
+        you the way to patch arriving messages, but also
+        places a restriction: until the callback exits, the
+        main event IPDB loop is blocked.
+
+        Normally, only "post" callbacks are required. But in
+        some specific cases "pre" also can be useful.
+
+        ...
+
+        The routine, `register_callback()`, takes two arguments:
+        1. callback function
+        2. mode (optional, default="post")
+
+        The callback should be a routine, that accepts three
+        arguments::
+
+            cb(ipdb, msg, action)
+
+        1. ipdb is a reference to IPDB instance, that invokes
+           the callback.
+        2. msg is a message arrived
+        3. action is just a msg['event'] field
+
+        E.g., to work on a new interface, you should catch
+        action == 'RTM_NEWLINK' and with the interface index
+        (arrived in msg['index']) get it from IPDB::
+
+            index = msg['index']
+            interface = ipdb.interfaces[index]
+        '''
+        if mode == 'post':
+            self._post_callbacks.append(callback)
+        elif mode == 'pre':
+            self._pre_callbacks.append(callback)
+
+    def unregister_callback(self, callback, mode='post'):
+        if mode == 'post':
+            cbchain = self._post_callbacks
+        elif mode == 'pre':
+            cbchain = self._pre_callbacks
+        else:
+            raise KeyError('Unknown callback mode')
+        for cb in tuple(cbchain):
             if callback == cb:
-                self._callbacks.pop(self._callbacks.index(cb))
+                cbchain.pop(cbchain.index(cb))
 
     def release(self):
         '''
         Shutdown monitoring thread and release iproute.
         '''
-        self._stop = True
-        self.nl.get_links()
-        self.nl.release()
-        self._mthread.join()
+        with self.exclusive:
+            self._stop = True
+            self.nl.get_links()
+            self.nl.release()
+            self._mthread.join()
 
     def create(self, kind, ifname, **kwarg):
         '''
@@ -892,7 +1158,7 @@ class IPDB(object):
 
         FIXME: this should be documented.
         '''
-        i = self.iclass(nl=self.nl, ipdb=self, mode='snapshot')
+        i = self.iclass(ipdb=self, mode='snapshot')
         i['kind'] = kind
         i['index'] = kwarg.get('index', 0)
         i['ifname'] = ifname
@@ -909,9 +1175,10 @@ class IPDB(object):
     def wait_interface(self, action='RTM_NEWLINK', **kwarg):
         event = threading.Event()
 
-        def cb(self, port, _action):
+        def cb(self, msg, _action):
             if _action != action:
                 return
+            port = self.interfaces[msg['index']]
             for key in kwarg:
                 if port.get(key, None) != kwarg[key]:
                     return
@@ -927,6 +1194,10 @@ class IPDB(object):
         # unregister callback
         self.unregister_callback(cb)
 
+    def update_routes(self, routes):
+        for msg in routes:
+            self.routes.load(msg)
+
     def update_links(self, links):
         '''
         Rebuild links index from list of RTM_NEWLINK messages.
@@ -938,7 +1209,7 @@ class IPDB(object):
                 self.by_index[dev['index']] = \
                 self.interfaces[dev['index']] = \
                 self.interfaces.get(dev.get_attr('IFLA_IFNAME'), None) or \
-                self.iclass(nl=self.nl, ipdb=self, mode=self.mode)
+                self.iclass(ipdb=self)
             i.load(dev)
             self.interfaces[i['ifname']] = \
                 self.by_name[i['ifname']] = i
@@ -1030,6 +1301,14 @@ class IPDB(object):
             except:
                 continue
             for msg in messages:
+                # run pre-callbacks
+                # NOTE: pre-callbacks are synchronous
+                for cb in self._pre_callbacks:
+                    try:
+                        cb(self, msg, msg['event'])
+                    except:
+                        pass
+
                 index = msg.get('index', None)
                 if msg.get('event', None) == 'RTM_NEWLINK':
                     if index in self.interfaces:
@@ -1069,15 +1348,28 @@ class IPDB(object):
                     self.update_addr([msg], 'add')
                 elif msg.get('event', None) == 'RTM_DELADDR':
                     self.update_addr([msg], 'remove')
+                elif msg.get('event', None) == 'RTM_NEWROUTE':
+                    self.update_routes([msg])
+                elif msg.get('event', None) == 'RTM_DELROUTE':
+                    table = msg.get('table', 254)
+                    dst = msg.get_attr('RTA_DST', False)
+                    if not dst:
+                        key = 'default'
+                    else:
+                        key = '%s/%s' % (dst, msg.get('dst_len', 0))
+                    try:
+                        route = self.routes.tables[table][key]
+                        del self.routes.tables[table][key]
+                        route.sync()
+                    except KeyError:
+                        pass
 
-                # run callbacks
-                for cb in self._callbacks:
+                # run post-callbacks
+                # NOTE: post-callbacks are asynchronous
+                for cb in self._post_callbacks:
                     t = threading.Thread(name="callback %s" % (id(cb)),
                                          target=cb,
-                                         args=(self,
-                                               self.interfaces.get(index,
-                                                                   None),
-                                               msg['event']))
+                                         args=(self, msg, msg['event']))
                     t.start()
                     self._cb_threads.add(t)
 
